define bi_nodejs::forever (
    $config_file = 'config.js',
    $config_content = undef,
    $script_path = undef,
    $app_name = $title,
    $ensure = undef,
    $port = undef,
    $owner = undef,
    $group = undef,
    $log_file = undef,
    $out_file = undef,
    $err_file = undef,
    $append = true,
    $pid_file = undef,
    $source_dir = undef,
    $working_dir = undef,
    $min_uptime = undef,
    $spin_sleep_time = undef,
    $colors = false,
    $plain = false,
    $debug = false,
    $verbose = false,
    $silent = false,
    $watch_dir = undef,
    $watch_ignore = undef,
    $kill_signal = undef,
    $max = undef,
    $path = undef,
    $fifo = undef,
    $command = undef,
    $number = undef,
    $watch = undef,
) {
    require ::bi_grep       # for init.d script
    require ::bi_coreutils  # for init.d script
    
    anchor { "bi_nodejs::forever::${app_name}::start": }
    anchor { "bi_nodejs::forever::${app_name}::end": }
    
    if $config_file == undef or !is_string($config_file) { 
        fail("'config_file' must be defined") 
    }
    
    if $config_content == undef or !is_string($config_content) { 
        fail("'config_content' must be defined") 
    }
    
    if $script_path == undef or !is_string($script_path) { 
        fail("'script_path' must be defined") 
    } else {
        validate_absolute_path($script_path)
    }
    
    if $app_name == undef or !is_string($app_name) { 
        fail("'app_name' must be defined") 
    } else {
        $arg_uid = "--uid $app_name"
    }
    
    if $ensure == undef or !is_string($ensure) { 
        fail("'ensure' must be defined") 
    } else {
        validate_re($ensure, 'running|stopped', "'ensure' must be one of 'running', 'stopped'")
    }
    
    if $port == undef or !is_numeric($port) { 
        fail("'port' must be defined and an integer") 
    }
    
    if $owner == undef or !is_string($owner) { 
        fail("'owner' must be defined") 
    }
    
    if $group == undef or !is_string($group) { 
        fail("'group' must be defined") 
    }
    
    if $max != undef {
        validate_integer($max)
        $arg_max = "-m $max"
    } else {
        $arg_max = undef
    }
    
    if $log_file != undef {
        validate_absolute_path(dirname($log_file))
        $log_dir = dirname($log_file)
        $arg_log_file = "-l $log_file"
    } else {
        $arg_log_file = "-l /var/log/$app_name/forever.log"
        $log_dir = "/var/log/$app_name"
    }    
    
    if $out_file != undef { 
        validate_absolute_path(dirname($out_file))
        $arg_out_file = "-o $out_file"
    } else {
        $arg_out_file = "-o /var/log/$app_name/out.log"
    }
    
    if $err_file != undef { 
        validate_absolute_path(dirname($err_file))
        $arg_err_file = "-e $err_file"
    } else {
        $arg_err_file = "-e /var/log/$app_name/err.log"
    }

    if $path != undef { 
        validate_absolute_path($path)
        $arg_path = "-p $path"
    } else {
        $arg_path = undef
    }

    if $command != undef {
        if is_string($command) {
            $arg_command = '-c $command'
        } else {
            fail("'command' must be a string")   
        }
    } else {
        $arg_command = undef
    }
    
    if $append != undef {
        notice("append: $append")
        validate_bool($append)
        $arg_append = $append ? {
            true => "--append",
            default => undef,
        }
    } else {
        $arg_append = undef
    }
    
    if $fifo != undef {
        validate_bool($fifo)
        $arg_fifo = $fifo ? {
            true => "--fifo",
            default => undef,
        }
    } else {
        $arg_fifo = undef
    }
 
    if $number != undef {
        validate_integer($number)
        $arg_number = "--number $number"
    } else {
        $arg_number = undef
    }
 
    if $pid_file != undef {
       validate_absolute_path(dirname($pid_file))
       $arg_pid_file = "--pidFile $pid_file"
       $real_pid_file = $pid_file
    } else {
        $real_pid_file = "/var/run/$app_name.pid"
        $arg_pid_file = "--pidFile $real_pid_file"
    }
 
    if $source_dir != undef { 
        validate_absolute_path($source_dir)
        $arg_source_dir = "--sourceDir $source_dir"
    } else {
        $arg_source_dir = undef
    }
    
    if $working_dir != undef { 
        validate_absolute_path($working_dir)
        $arg_working_dir = "--workingDir $working_dir"
    } else {
        $arg_working_dir = undef
    }
    
    if $min_uptime != undef {
        validate_integer($min_uptime)
        $arg_min_uptime = "--minUptime $min_uptime"
    } else {
        $arg_min_uptime = undef
    }
    
    if $spin_sleep_time != undef {
        validate_integer($spin_sleep_time)
        $arg_spin_sleep_time = "--spinSleepTime $spin_sleep_time"
    } else {
        $arg_spin_sleep_time = undef
    }
    
    if $colors != undef {
        validate_bool($colors)
        $arg_colors = $colors ? {
            true => "--colors",
            default => undef,
        }
    } else {
        $arg_colors = undef
    }

    if $plain != undef {
        validate_bool($plain)
        $arg_plain = $plain ? {
            true => "--plain",
            default => undef,
        }
    } else {
        $arg_plain = undef
    }
    
    if $colors and $plain {
        fail("'colors' and 'plain' are mutually exclusive") 
    }
    
    if $debug != undef {
        validate_bool($debug)
        $arg_debug = $debug ? {
            true => "--debug",
            default => undef,
        }
    } else {
        $arg_debug = undef
    }

    if $verbose != undef {
        validate_bool($verbose)
        $arg_verbose = $verbose ? {
            true => "--verbose",
            default => undef,
        }
    } else {
        $arg_verbose = undef
    }
    
    if $silent != undef {
        validate_bool($silent)
        $arg_silent = $silent ? {
            true => "--silent",
            default => undef,
        }
    } else {
        $arg_silent = undef
    }
    
    if $watch != undef {
        validate_bool($watch)
        $arg_watch = $watch ? {
            true => "--watch",
            default => undef,
        }
    } else {
        $arg_watch = undef
    }
    
    if $watch_dir != undef {
        validate_absolute_path($watch_dir)
        $arg_watch_dir = "--watchDirectory $watch_dir"
    } else {
        $arg_watch_dir = undef
    }
    
    if $watch_ignore != undef {
        validate_bool($watch_ignore)
        $arg_watch_ignore = $watch_ignore ? {
            true => "--watchIgnore",
            default => undef,
        }
    } else {
        $arg_watch_ignore = undef
    }
    
    if $kill_signal != undef { 
        if is_string($kill_signal) { 
             $arg_kill_signal = "--killSignal $kill_signal"
        } else {
            fail("'kill_signal' must a string")
        } 
    } else {
        $arg_kill_signal = undef
    }
    
    $args = join(delete_undef_values([ 
        $arg_log_file,
        $arg_out_file,
        $arg_err_file,
        $arg_append,
        $arg_pid_file,
        $arg_source_dir,
        $arg_working_dir,
        $arg_min_uptime,
        $arg_spin_sleep_time,
        $arg_colors,
        $arg_plain,
        $arg_debug,
        $arg_verbose,
        $arg_silent,
        $arg_watch_dir,
        $arg_watch_ignore,
        $arg_kill_signal,
        $arg_uid,
        $arg_max,
        $arg_path,
        $arg_fifo,
        $arg_command,
        $arg_number,
        $arg_watch,
    ]), " ")
    
    $start_command = "forever start $args $script_path"
    $stop_command = "forever stop $app_name"
    
    file { "/etc/init.d/$app_name":
        owner => 'root',
        group => 'root',
        mode => '755',
        content => template("$module_name/etc/init.d/forever.bash.erb"),
        require => [
            Class[::bi_grep],
            Class[::bi_coreutils],  
        ],
    }
    
    file { $log_dir:
        ensure => 'directory',
        owner => $owner,
        group => $group,
        mode => '755',
    }
    
    file { $real_pid_file:
        ensure => 'present',
        owner => $owner,
        group => $group,
        mode => '755',
        content => "",
    }
    
    file { "$working_dir/$config_file":
        ensure => 'present',
        owner => $owner,
        group => $group,
        mode => '755',
        content => $config_content,
    }
    
    Anchor["bi_nodejs::forever::${app_name}::start"] ->
    #Class[::bi_grep] ->
    #Class[::bi_coreutils] ->
    File["/etc/init.d/$app_name"] ->
    File[$log_dir] ->
    File[$real_pid_file] ->
    File["$working_dir/$config_file"] ->
    Anchor["bi_nodejs::forever::${app_name}::end"]
    
}
